from pathlib import Path
import os
import shutil
import uuid

import h5py
import numpy as np
import pytest

from bmlab.session import Session
from bmlab.models.calibration_model import FitSet, RayleighFit
from bmlab.models.extraction_model import CircleFit
from bmlab.controllers import CalibrationController, ExtractionController


@pytest.fixture()
def tmp_dir():
    tmp_dir = 'tmp' + str(uuid.uuid4())
    os.mkdir(tmp_dir)
    os.chdir(tmp_dir)
    yield tmp_dir
    os.chdir('..')
    try:
        shutil.rmtree(tmp_dir)
    # Windows sometimes does not correctly close the HDF file. Then we have a
    # file access conflict.
    except Exception as e:
        print(e)


@pytest.fixture()
def session_file(tmp_dir):

    session = Session.get_instance()

    shutil.copy(data_file_path('Water.h5'), Path.cwd() / 'Water.h5')

    session.set_file('Water.h5')

    session.set_reflection(vertically=True, horizontally=False)

    session.set_current_repetition('0')

    ec = ExtractionController()
    cc = CalibrationController()

    em = session.extraction_model()
    cm = session.calibration_model()

    img = session.get_payload_image('0', 0)
    em.set_image_shape(img.shape)

    points = [(100, 290), (145, 255), (290, 110)]
    for calib_key in session.get_calib_keys():
        for p in points:
            ec.add_point(calib_key, p)

        assert em.get_arc_by_calib_key(calib_key) != np.empty(0)
        assert cm.get_spectra(calib_key) is None

        cc.extract_spectra(calib_key)

    session.save()

    session.clear()

    yield 'Water.session.h5'


def data_file_path(file_name):
    return Path(__file__).parent / 'data' / file_name


def test_serialize_session(tmp_dir):
    session = Session.get_instance()

    shutil.copy(data_file_path('Water.h5'), Path.cwd() / 'Water.h5')

    session.set_file('Water.h5')

    session.set_reflection(vertically=True, horizontally=False)

    session.set_current_repetition('0')

    ec = ExtractionController()
    cc = CalibrationController()

    em = session.extraction_model()
    cm = session.calibration_model()

    points = [(100, 290), (145, 255), (290, 110)]
    for calib_key in session.get_calib_keys():
        for p in points:
            ec.add_point(calib_key, p)

        assert em.get_arc_by_calib_key(calib_key) != np.empty(0)
        assert cm.get_spectra(calib_key) is None

        cc.extract_spectra(calib_key)

    session.save()

    with h5py.File('Water.session.h5', 'r') as f:
        assert 'session/extraction_models/0/points/1' in f
        assert f.attrs['version'].startswith('bmlab')

    session.clear()


def test_deserialize_session_file(session_file):

    session = Session.get_instance()
    session.set_file('Water.h5')

    em = session.extraction_model()
    cm = session.calibration_model()
    assert em
    assert em.calib_times['2'] == 62.542
    assert len(cm.spectra['1']) > 0
    assert len(em.positions['2']) > 0
    assert em.positions_interpolation is not None

    session.clear()


def test_serialize_fitset(tmp_dir):

    fit_set = FitSet()
    fit = RayleighFit('1', 3, 4, 11., 12., 13., 14.)
    fit_set.add_fit(fit)

    with h5py.File('tmpsession.h5', 'w') as f:
        fit_set.serialize(f, 'fits')

    with h5py.File('tmpsession.h5', 'r') as f:
        actual = FitSet.deserialize(f['fits'])

    actual_fit = actual.get_fit('1', 3, 4)
    expected_fit = fit_set.get_fit('1', 3, 4)
    assert actual_fit.w0 == expected_fit.w0
    assert actual_fit.fwhm == expected_fit.fwhm
    assert actual_fit.intensity == expected_fit.intensity
    assert actual_fit.offset == expected_fit.offset


def test_de_serialize_CircleFit(tmp_dir):

    cf = CircleFit(center=(1., 2.), radius=3.)

    with h5py.File(str(tmp_dir) + 'abc.h5', 'w') as f:
        cf.serialize(f, 'circle')

        assert isinstance(f['circle'], h5py.Group)
        assert f['circle'].attrs['type'] == \
               'bmlab.models.extraction_model.CircleFit'

    with h5py.File(str(tmp_dir) + 'abc.h5', 'r') as f:
        cf = CircleFit.deserialize(f['circle'])

        np.testing.assert_array_equal(cf.center, (1., 2.))
        assert cf.radius == 3.



import numpy as np


def regions_merge_add_region(regions, region):
    regions_fused = False

    # check if the selected regions overlap
    for i, saved_region in enumerate(regions):
        if (np.min(region) < np.max(saved_region)
                and (np.max(region) > np.min(saved_region))):
            # fuse overlapping regions
            regions[i] = (
                np.min([region, saved_region]),
                np.max([region, saved_region]))
            regions_fused = True

    if not regions_fused:
        regions.append(region)


def regions_check_order(regions):
    """
    This function checks that the tuples in the regions array
    are sorted based on their center

    Parameters
    ----------
    regions: array containing regions

    """
    regions.sort(key=np.mean)



from bmlab.serializer import Serializer


class Orientation(Serializer):

    def __init__(self, rotation=0,
                 reflection={'vertically': False, 'horizontally': False}):
        self.rotation = rotation
        self.reflection = reflection

    def set_rotation(self, num_rots):
        self.rotation = num_rots

    def set_reflection(self, **kwargs):
        axes = ['vertically', 'horizontally']
        for a in axes:
            if a in kwargs:
                self.reflection[a] = kwargs[a]

    def apply(self, img):
        from bmlab.image import set_orientation
        return set_orientation(img, self.rotation,
                               self.reflection['vertically'],
                               self.reflection['horizontally'])



import os
import numpy as np
import csv
from PIL import Image
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import Normalize
from matplotlib.backends.backend_pdf import PdfPages

from bmlab import Session


class BrillouinExport(object):

    def __init__(self, evc):
        self.session = Session.get_instance()
        self.file = self.session.file
        self.mode = 'Brillouin'
        self.evc = evc
        return

    def export(self, configuration):
        if not self.file:
            return

        config = configuration['brillouin']
        if not config['export']:
            return

        brillouin_repetitions = self.file.repetition_keys()

        for brillouin_repetition in brillouin_repetitions:
            self.session.set_current_repetition(brillouin_repetition)

            # Get a list of all parameters available
            parameters = self.session.evaluation_model().get_parameter_keys()

            for parameter_key in config['parameters']:

                if parameter_key not in parameters:
                    continue

                # Get the data to plot
                data, positions, dimensionality, labels =\
                    self.evc.get_data(parameter_key)

                # Subtract the mean value of the positions,
                # so they are centered around zero
                for position in positions:
                    position -= np.nanmean(position)

                # This only works for two-dim data!
                if dimensionality != 2:
                    continue

                # Create data necessary to correctly slice the data
                dslice = [slice(None) if dim > 1 else 0 for dim in data.shape]
                idx = [idx for idx, dim in enumerate(data.shape) if dim > 1]

                # We rotate the array, so the x axis is shown as the
                # horizontal axis
                image_map = data[tuple(dslice)]
                image_map = np.rot90(image_map)
                extent = np.nanmin(positions[idx[0]][tuple(dslice)]), \
                    np.nanmax(positions[idx[0]][tuple(dslice)]), \
                    np.nanmin(positions[idx[1]][tuple(dslice)]), \
                    np.nanmax(positions[idx[1]][tuple(dslice)])

                # Actually plot the data
                fig = plt.figure()

                plot = fig.add_subplot(111)

                ims = plt.imshow(
                    image_map, interpolation='nearest',
                    extent=extent
                )
                plot.set_xlabel(labels[idx[0]])
                plot.set_ylabel(labels[idx[1]])
                cb_label = parameters[parameter_key]['symbol'] + \
                    ' [' + parameters[parameter_key]['unit'] + ']'
                colorbar = fig.colorbar(ims)
                colorbar.ax.set_title(cb_label)
                plot.axis('scaled')
                plot.set_xlim(
                    np.nanmin(positions[idx[0]][tuple(dslice)]),
                    np.nanmax(positions[idx[0]][tuple(dslice)])
                )
                plot.set_ylim(
                    np.nanmin(positions[idx[1]][tuple(dslice)]),
                    np.nanmax(positions[idx[1]][tuple(dslice)])
                )

                # Set the colormap limits to min/max
                # or provided boundaries
                # Minimum limit
                try:
                    value_min = float(config[parameter_key]['cax'][0])
                except BaseException:
                    value_min = np.nanmin(image_map)
                try:
                    value_max = float(config[parameter_key]['cax'][1])
                except BaseException:
                    value_max = np.nanmax(image_map)
                ims.set_clim(value_min, value_max)

                # Export plot
                if self.file.path.parent.name == 'RawData':
                    path = self.file.path.parents[1] / 'Plots' / 'WithAxis'
                else:
                    path = self.file.path.parent
                if not os.path.exists(path):
                    os.makedirs(path, exist_ok=True)

                # Export as PDF
                pdf_path = path / f"{self.file.path.stem}" \
                                  f"_BMrep{brillouin_repetition}" \
                                  f"_{parameter_key}" \
                                  ".pdf"
                with PdfPages(pdf_path) as pdf:
                    pdf.savefig()

                # Export as PNG
                png_path = path / f"{self.file.path.stem}" \
                                  f"_BMrep{brillouin_repetition}" \
                                  f"_{parameter_key}" \
                                  ".png"
                plt.savefig(png_path)

                plt.close(fig)

                # Export as bare image without axes
                if self.file.path.parent.name == 'RawData':
                    path = self.file.path.parents[1] / 'Plots' / 'Bare'
                else:
                    path = self.file.path.parent
                if not os.path.exists(path):
                    os.makedirs(path, exist_ok=True)

                # Convert indexed data into rgba map
                rgba = cm.ScalarMappable(
                    norm=Normalize(vmin=value_min, vmax=value_max),
                    cmap=cm.viridis
                ).to_rgba(image_map)

                filename = path / f"{self.file.path.stem}" \
                                  f"_BMrep{brillouin_repetition}" \
                                  f"_{parameter_key}" \
                                  ".png"
                image = Image.fromarray((255 * rgba).astype(np.ubyte))
                image.save(filename)

                # Export as TIFF files
                filename = path / f"{self.file.path.stem}" \
                                  f"_BMrep{brillouin_repetition}" \
                                  f"_{parameter_key}" \
                                  ".tiff"
                image = Image.fromarray(10000 * image_map)
                image.save(filename)

                # Export data as CSV file
                if self.file.path.parent.name == 'RawData':
                    csv_path = self.file.path.parents[1] / 'Export'
                else:
                    csv_path = self.file.path.parent
                if not os.path.exists(csv_path):
                    os.makedirs(csv_path, exist_ok=True)
                csv_filename = csv_path / f"{self.file.path.stem}" \
                                          f"_BMrep{brillouin_repetition}" \
                                          f"_{parameter_key}" \
                                          ".csv"
                with open(csv_filename, 'w', newline='') as csvfile:
                    csv_writer = csv.writer(
                        csvfile, delimiter=',',
                        quotechar='|', quoting=csv.QUOTE_MINIMAL)
                    csv_writer.writerow(['Brillouin shift [GHz]'])
                    csv_writer.writerow([])
                    csv_writer.writerows(data[tuple(dslice)])



import os
import numpy as np
from skimage import transform
from PIL import Image

from bmlab import Session


class FluorescenceExport(object):

    def __init__(self):
        self.session = Session.get_instance()
        self.file = self.session.file
        self.mode = 'Fluorescence'
        return

    def export(self, configuration):
        if not self.file:
            return

        config = configuration['fluorescence']
        if not config['export']:
            return

        fluorescence_repetitions = self.file.repetition_keys(self.mode)
        brillouin_repetitions = self.file.repetition_keys()

        # Loop over all fluorescence repetitions
        for fluorescence_repetition in fluorescence_repetitions:
            # Get the repetition
            repetition = self.file.get_repetition(
                fluorescence_repetition, self.mode)
            # Get the keys for all images in this repetition
            image_keys = repetition.payload.image_keys()
            # Get the scale calibration
            scale_calibration = repetition.payload.get_scale_calibration()
            # Create the transform matrix for the affine transformation
            n = np.linalg.norm(np.array(scale_calibration['micrometerToPixX']))
            tmatrix = np.matrix([
                [-1 * scale_calibration['micrometerToPixY'][1],
                 -1 * scale_calibration['micrometerToPixY'][0], 0],
                [-1 * scale_calibration['micrometerToPixX'][1],
                 -1 * scale_calibration['micrometerToPixX'][0], 0],
                [0, 0, n]
            ]) / n

            # Loop over all images in this repetition
            for image_key in image_keys:
                channel = repetition.payload.get_channel(image_key)
                img_data = repetition.payload.get_image(image_key)
                # Average all images acquired
                img_data = np.nanmean(img_data, axis=0).astype(np.ubyte)

                # Get the region of interest of the repetition
                roi = repetition.payload.get_ROI(image_key)

                # Create an x-y grid for the image with positions in µm
                x_pix, y_pix = np.meshgrid(
                    np.arange(roi['width_physical']) + roi['left'],
                    np.flip(np.arange(roi['height_physical']) + roi['bottom'])
                )
                x_pix = x_pix - scale_calibration['origin'][0]
                y_pix = y_pix - scale_calibration['origin'][1]

                x_mm =\
                    x_pix * scale_calibration['pixToMicrometerX'][0] +\
                    y_pix * scale_calibration['pixToMicrometerY'][0] +\
                    scale_calibration['positionStage'][0]
                y_mm =\
                    x_pix * scale_calibration['pixToMicrometerX'][1] +\
                    y_pix * scale_calibration['pixToMicrometerY'][1] +\
                    scale_calibration['positionStage'][1]

                # Construct export path and create it if necessary
                if self.file.path.parent.name == 'RawData':
                    path = self.file.path.parents[1] / 'Plots'
                else:
                    path = self.file.path.parent
                if not os.path.exists(path):
                    os.mkdir(path)
                filename = path / f"{self.file.path.stem}" \
                                  f"_FLrep{fluorescence_repetition}" \
                                  f"_channel{channel}.png"
                image = Image.fromarray(img_data)
                if channel.casefold() == 'red':
                    blank = Image.new("L", image.size)
                    image = Image.merge("RGB", (image, blank, blank))

                if channel.casefold() == 'green':
                    blank = Image.new("L", image.size)
                    image = Image.merge("RGB", (blank, image, blank))

                if channel.casefold() == 'blue':
                    blank = Image.new("L", image.size)
                    image = Image.merge("RGB", (blank, blank, image))

                image.save(filename)

                # Also export the image with axes parallel to the stage

                # Create translation matrix to move image back to ROI
                corners = [[0, image.size[0]-1, 0, image.size[0]-1],
                           [0, 0, image.size[1]-1, image.size[1]-1],
                           [1, 1, 1, 1]]
                corners_warped = tmatrix * corners

                # Necessary translation
                dx = corners_warped[0, :].min()
                dy = corners_warped[1, :].min()
                # New shape
                sx = corners_warped[0, :].max()\
                    - corners_warped[0, :].min()
                sy = corners_warped[1, :].max()\
                    - corners_warped[1, :].min()

                translate = np.matrix([
                    [1, 0, -dx],
                    [0, 1, -dy],
                    [0, 0, 1]
                ])

                tform = transform.AffineTransform(
                    matrix=np.linalg.inv(translate * tmatrix))
                shape = (int(np.ceil(sy)), int(np.ceil(sx)))

                # Warp the images and positions to align with a
                # standard x-y coordinate system
                image_data_warped = transform.warp(
                    img_data,
                    tform, output_shape=shape, cval=np.nan)
                x_mm_warped = transform.warp(
                    x_mm,
                    tform, output_shape=shape, cval=np.nan)
                y_mm_warped = transform.warp(
                    y_mm,
                    tform, output_shape=shape, cval=np.nan)

                # Export image with proper alpha channel
                image_warped = Image.fromarray(
                    (255 * image_data_warped).astype(np.ubyte))
                if channel.casefold() == 'red':
                    blank = Image.new("L", image_warped.size)
                    image_warped =\
                        Image.merge("RGB", (image_warped, blank, blank))

                if channel.casefold() == 'green':
                    blank = Image.new("L", image_warped.size)
                    image_warped =\
                        Image.merge("RGB", (blank, image_warped, blank))

                if channel.casefold() == 'blue':
                    blank = Image.new("L", image_warped.size)
                    image_warped =\
                        Image.merge("RGB", (blank, blank, image_warped))
                image_alpha = Image.fromarray(
                    (255 * np.logical_not(
                        np.isnan(image_data_warped))).astype(np.ubyte))
                image_warped.putalpha(image_alpha)

                filename = path / f"{self.file.path.stem}" \
                                  f"_FLrep{fluorescence_repetition}" \
                                  f"_channel{channel}_aligned.png"
                image_warped.save(filename)

                # Export the images with the ROI of the Brillouin
                # measurements
                for brillouin_repetition in brillouin_repetitions:
                    # Get the repetition
                    repetition_bm = self.file.get_repetition(
                        brillouin_repetition)
                    # Read the Brillouin positions
                    positions = repetition_bm.payload.positions
                    x_min = np.nanmin(positions['x'])
                    x_max = np.nanmax(positions['x'])
                    y_min = np.nanmin(positions['y'])
                    y_max = np.nanmax(positions['y'])

                    # We are only interested in x-y-maps
                    if not x_min < x_max or not y_min < y_max:
                        continue

                    # Find the indices delimiting the Brillouin ROI
                    idx_mask = (x_mm_warped >= x_min) &\
                               (x_mm_warped <= x_max) &\
                               (y_mm_warped >= y_min) &\
                               (y_mm_warped <= y_max)
                    idx = np.nonzero(idx_mask)

                    # Crop the Fluorescence image to the Brillouin ROI
                    image_warped_bm = image_warped.crop(
                        (
                            np.min(idx[1]),
                            np.min(idx[0]),
                            np.max(idx[1]),
                            np.max(idx[0])
                        )
                    )

                    filename = path / f"{self.file.path.stem}" \
                                      f"_FLrep{fluorescence_repetition}" \
                                      f"_channel{channel}" \
                                      f"_BMrep{brillouin_repetition}.png"
                    image_warped_bm.save(filename)



"""
The serializer module contains functions that allow
serialization / deserialization of custom objects to/from HDF files.
"""

import importlib
import inspect
import re
import logging
import numbers
import builtins

import h5py


logger = logging.getLogger(__name__)


class Serializer(object):

    def serialize(self, parent, name, skip=[]):
        group = parent.create_group(name)

        group.attrs['type'] = '%s.%s' % (self.__class__.__module__,
                                         self.__class__.__name__)

        for var_name, var_value in self.__dict__.items():

            if callable(var_value):
                continue

            if var_name in skip:
                continue

            self.do_serialize(group, var_value, var_name)

    def serialize_dict(self, parent, dic, name):
        group = parent.create_group(name)
        for key, value in dic.items():
            self.serialize(group, key)

    def serialize_array(self, parent, array, name):
        parent.create_dataset(name, data=array)

    def serialize_string(self, parent, string, name):
        ds = parent.create_dataset(name, data=string,
                                   dtype=h5py.string_dtype())
        ds.attrs['type'] = 'builtins.str'

    def serialize_number(self, parent, number, name):
        parent.create_dataset(name, data=number)

    def do_serialize(self, group, value, name):
        if isinstance(value, dict):
            wrapped_dict = SerializableDict(value)
            wrapped_dict.serialize(group, name)
        elif isinstance(value, list):
            wrapped_list = SerializableList(value)
            wrapped_list.serialize(group, name)
        elif isinstance(value, tuple):
            wrapped_list = SerializableTuple(value)
            wrapped_list.serialize(group, name)
        elif isinstance(value, Serializer):
            value.serialize(group, name)
        elif is_list_like(value) and not isinstance(value, str):
            self.serialize_array(group, value, name)
        elif is_scalar(value):
            self.serialize_number(group, value, name)
        elif isinstance(value, str):
            self.serialize_string(group, value, name)
        elif value is None:
            pass
        else:
            raise Exception('Cannot serialize variable %s' % name)

    @classmethod
    def deserialize(cls, group):
        class_ = group.attrs.get('type')

        if class_ == 'builtins.list':
            instance = list()
            instance_handle = instance
            num_items = len(group.keys())
            for idx in range(num_items):
                value = group[str(idx)]
                instance.append(None)
                cls.do_deserialize(instance_handle, value, -1)
            return instance
        elif class_ == 'builtins.tuple':
            instance = list()
            instance_handle = instance
            num_items = len(group.keys())
            for idx in range(num_items):
                value = group[str(idx)]
                instance.append(None)
                cls.do_deserialize(instance_handle, value, -1)
            return tuple(instance)

        if class_ == 'builtins.dict':
            instance = dict()
            instance_handle = instance
        else:
            instance = _init_raw_object(class_)
            instance_handle = instance.__dict__

        for var_name, var_value in group.items():
            cls.do_deserialize(instance_handle, var_value, var_name)

        if isinstance(instance, Serializer):
            instance.post_deserialize()

        return instance

    @classmethod
    def do_deserialize(cls, instance_handle, var_value, var_name):
        if isinstance(var_value, h5py.Dataset):
            if var_value.shape == ():
                item = var_value[...].item()
                if isinstance(item, bytes):
                    item = item.decode('utf-8')
                instance_handle[var_name] = item
            else:
                instance_handle[var_name] = var_value[...]
        elif isinstance(var_value, h5py.Group):
            instance_handle[var_name] = Serializer.deserialize(var_value)
        else:
            raise Exception('Cannot deserialize object %s' % var_name)

    def post_deserialize(self):
        """
        Override this method in derived classes if further initialization
        is necessary after deserialization from HDF.
        """
        pass


class SerializableTuple(Serializer, builtins.tuple):

    def __init__(self, pure_tuple):
        self.pure_tuple = pure_tuple

    def serialize(self, parent, name):
        group = parent.create_group(name)
        group.attrs['type'] = 'builtins.tuple'
        for idx, value in enumerate(self.pure_tuple):
            self.do_serialize(group, value, str(idx))


class SerializableList(Serializer, builtins.list):

    def __init__(self, pure_list):
        self.pure_list = pure_list

    def serialize(self, parent, name):
        group = parent.create_group(name)
        group.attrs['type'] = 'builtins.list'
        for idx, value in enumerate(self.pure_list):
            self.do_serialize(group, value, str(idx))


class SerializableDict(Serializer, builtins.dict):

    def __init__(self, pure_dict):
        self.pure_dict = pure_dict

    def serialize(self, parent, name):
        group = parent.create_group(name)

        group.attrs['type'] = 'builtins.dict'

        for key, value in self.pure_dict.items():
            if callable(value):
                continue
            self.do_serialize(group, value, key)


def is_scalar(value):
    return isinstance(value, str) or isinstance(value, numbers.Number)


def is_list_like(value):
    return hasattr(value, '__len__') and not isinstance(value, dict)


def _init_raw_object(full_class_name):
    """
    Create an object without calling its constructor.

    Parameters
    ----------
    full_class_name: str
        fully qualified name of the class (package.classname)

    Returns
    -------
    Instance of the requested class with no properties
    initialized.
    """
    class_ = _class_from_full_class_name(full_class_name)
    return class_.__new__(class_)


def _class_from_full_class_name(full_class_name):

    pattern = re.compile('(.+)[.]([^.]+)')
    match = pattern.match(full_class_name)
    module_name = match[1]
    class_name = match[2]
    module = importlib.import_module(module_name)
    return _class_in_module(module, class_name)


def _class_in_module(module, class_name):

    for name, class_ in inspect.getmembers(module, inspect.isclass):
        if name == class_name:
            return class_



from bmlab.models.calibration_model import CalibrationModel,\
    FitSet, BrillouinFit

import numpy as np


def test_calibration_model_add_brillouin_region():
    cm = CalibrationModel()

    brillouin_region = cm.get_brillouin_regions(0)
    assert brillouin_region == []

    cm.add_brillouin_region(0, [0.9, 3])
    cm.add_brillouin_region(0, (2, 5))

    brillouin_region_0 = cm.get_brillouin_regions(0)

    assert brillouin_region_0 == [(1, 5)]

    cm.add_brillouin_region(0, (7, 9))

    assert brillouin_region_0 == [(1, 5), (7, 9)]

    cm.add_brillouin_region(0, (2, 3))

    assert brillouin_region_0 == [(1, 5), (7, 9)]


def test_calibration_model_set_brillouin_region():
    cm = CalibrationModel()
    cm.add_brillouin_region(0, [0.9, 3])

    brillouin_region_0 = cm.get_brillouin_regions(0)

    assert brillouin_region_0 == [(1, 3)]

    cm.set_brillouin_region(0, 0, (2, 5))

    assert brillouin_region_0 == [(2, 5)]

    cm.set_brillouin_region(0, 1, (4, 7))

    assert brillouin_region_0 == [(2, 5), (4, 7)]

    cm.set_brillouin_region(0, 3, (8, 9))

    assert brillouin_region_0 == [(2, 5), (4, 7), (8, 9)]


def test_calibration_model_clear_brillouin_region():
    cm = CalibrationModel()
    cm.add_brillouin_region(0, [0.9, 3])

    brillouin_region_0 = cm.get_brillouin_regions(0)

    assert brillouin_region_0 == [(1, 3)]

    cm.clear_brillouin_regions(0)
    brillouin_region_0 = cm.get_brillouin_regions(0)

    assert brillouin_region_0 == []


def test_calibration_model_add_rayleigh_region():
    cm = CalibrationModel()

    rayleigh_region = cm.get_rayleigh_regions(0)
    assert rayleigh_region == []

    cm.add_rayleigh_region(0, [0.9, 3])
    cm.add_rayleigh_region(0, (2, 5))

    rayleigh_region_0 = cm.get_rayleigh_regions(0)

    assert rayleigh_region_0 == [(1, 5)]

    cm.add_rayleigh_region(0, (7, 9))

    assert rayleigh_region_0 == [(1, 5), (7, 9)]

    cm.add_rayleigh_region(0, (2, 3))

    assert rayleigh_region_0 == [(1, 5), (7, 9)]


def test_calibration_model_set_rayleigh_region():
    cm = CalibrationModel()
    cm.add_rayleigh_region(0, [0.9, 3])

    rayleigh_region_0 = cm.get_rayleigh_regions(0)

    assert rayleigh_region_0 == [(1, 3)]

    cm.set_rayleigh_region(0, 0, (2, 5))

    assert rayleigh_region_0 == [(2, 5)]

    cm.set_rayleigh_region(0, 1, (4, 7))

    assert rayleigh_region_0 == [(2, 5), (4, 7)]

    cm.set_rayleigh_region(0, 3, (8, 9))

    assert rayleigh_region_0 == [(2, 5), (4, 7), (8, 9)]


def test_calibration_model_clear_rayleigh_region():
    cm = CalibrationModel()
    cm.add_rayleigh_region(0, [0.9, 3])

    rayleigh_region_0 = cm.get_rayleigh_regions(0)

    assert rayleigh_region_0 == [(1, 3)]

    cm.clear_rayleigh_regions(0)
    rayleigh_region_0 = cm.get_rayleigh_regions(0)

    assert rayleigh_region_0 == []


def test_get_frequencies_by_calib_key():
    cm = CalibrationModel()

    frequencies0 = np.arange(100).reshape(1, -1)
    cm.set_frequencies('0', 0, list(frequencies0))

    frequencies1 = np.arange(100).reshape(1, -1) + 10
    cm.set_frequencies('1', 10, list(frequencies1))

    assert (cm.get_frequencies_by_calib_key('0') == frequencies0).all()
    assert (cm.get_frequencies_by_calib_key('1') == frequencies1).all()

    cm.clear_frequencies('0')
    assert cm.get_frequencies_by_calib_key('0') is None


def test_get_frequency_by_calib_key():
    cm = CalibrationModel()

    frequencies0 = list(np.arange(100).reshape(1, -1))
    cm.set_frequencies('0', 0, frequencies0)

    frequencies1 = list(np.arange(100).reshape(1, -1) + 10)
    cm.set_frequencies('1', 10, frequencies1)

    assert cm.get_frequency_by_calib_key(70.7, '0') == 70.7
    assert cm.get_frequency_by_calib_key(80, '1') == 90

    cm.clear_frequencies('0')
    assert cm.get_frequency_by_calib_key(70.7, '0') is None


def test_get_frequencies_by_time():
    cm = CalibrationModel()

    """
    Fill calibration model with values for calibrations
    """
    # Test that one calibration works
    frequencies0 = np.arange(100)
    cm.set_frequencies('0', 0, list(frequencies0.reshape(1, -1)))

    frequencies = cm.get_frequencies_by_time(0)
    np.testing.assert_allclose(frequencies, frequencies0, atol=1.E-8)

    frequencies = cm.get_frequencies_by_time(10)
    np.testing.assert_allclose(frequencies, frequencies0, atol=1.E-8)

    frequencies1 = np.arange(100) + 10
    cm.set_frequencies('1', 10, list(frequencies1.reshape(1, -1)))

    frequencies = cm.get_frequencies_by_time(0)
    np.testing.assert_allclose(frequencies, frequencies0, atol=1.E-8)

    frequencies = cm.get_frequencies_by_time(10)
    np.testing.assert_allclose(frequencies, frequencies1, atol=1.E-8)


def test_get_frequency_by_time():
    cm = CalibrationModel()

    """
    Fill calibration model with values for calibrations
    """
    # Test that one calibration works
    cm.set_frequencies('0', 0, list(np.arange(100).reshape(1, -1)))

    frequency = cm.get_frequency_by_time(0, 0)
    np.testing.assert_allclose(frequency, 0, atol=1.E-8)
    frequency = cm.get_frequency_by_time(10, 0)
    np.testing.assert_allclose(frequency, 0, atol=1.E-8)

    # Ensure we don't throw a hard error in case our requested value
    # is outside the interpolation range
    frequency = cm.get_frequency_by_time(0, -1)
    np.testing.assert_array_equal(frequency, np.nan)
    frequency = cm.get_frequency_by_time(0, 100)
    np.testing.assert_array_equal(frequency, np.nan)

    times = 10 * np.ones((5, 5, 5, 2, 2, 2))
    positions = np.zeros((5, 5, 5, 2, 2, 2))
    positions[:, :, :, 1, :, :] = 50
    frequencies = cm.get_frequency_by_time(times, positions)

    expected = np.zeros((5, 5, 5, 2, 2, 2))
    expected[:, :, :, 1, :, :] = 50
    np.testing.assert_allclose(frequencies, expected, atol=1.E-8)

    # Test that two calibrations work
    cm.set_frequencies('1', 10, list(np.arange(100).reshape(1, -1) + 10))

    frequency = cm.get_frequency_by_time(0, 0)
    np.testing.assert_allclose(frequency, 0, atol=1.E-8)
    frequency = cm.get_frequency_by_time(10, 0)
    np.testing.assert_allclose(frequency, 10, atol=1.E-8)

    # Ensure we don't throw a hard error in case our requested value
    # is outside the interpolation range
    frequency = cm.get_frequency_by_time(0, -1)
    np.testing.assert_array_equal(frequency, np.nan)
    frequency = cm.get_frequency_by_time(0, 100)
    np.testing.assert_array_equal(frequency, np.nan)

    times = 10 * np.ones((5, 5, 5, 2, 2, 2))
    positions = np.zeros((5, 5, 5, 2, 2, 2))
    positions[:, :, :, 1, :, :] = 50
    frequencies = cm.get_frequency_by_time(times, positions)

    expected = 10 * np.ones((5, 5, 5, 2, 2, 2))
    expected[:, :, :, 1, :, :] = 60
    np.testing.assert_allclose(frequencies, expected, atol=1.E-8)

    # Test that three calibrations work
    cm.set_frequencies('2', 30, list(np.arange(100).reshape(1, -1) + 30))

    frequency = cm.get_frequency_by_time(0, 0)
    np.testing.assert_allclose(frequency, 0, atol=1.E-8)

    frequency = cm.get_frequency_by_time(10, 0)
    np.testing.assert_allclose(frequency, 10, atol=1.E-8)

    frequency = cm.get_frequency_by_time(20, 0)
    np.testing.assert_allclose(frequency, 20, atol=1.E-8)

    frequency = cm.get_frequency_by_time(20, 11)
    np.testing.assert_allclose(frequency, 31, atol=1.E-8)

    # Test that the function also accepts arrays
    times = 10 * np.ones((5, 5, 5, 2, 2, 2))
    times[:, :, :, :, :, 1] = 20
    positions = np.zeros((5, 5, 5, 2, 2, 2))
    positions[:, :, :, 1, :, :] = 50
    frequencies = cm.get_frequency_by_time(times, positions)

    expected = 10 * np.ones((5, 5, 5, 2, 2, 2))
    expected[:, :, :, 0, :, 1] = 20
    expected[:, :, :, 1, :, 0] = 60
    expected[:, :, :, 1, :, 1] = 70
    np.testing.assert_allclose(frequencies, expected, atol=1.E-8)


def test_get_position_by_time():
    cm = CalibrationModel()

    frequencies0 = list(np.arange(100).reshape(1, -1))
    cm.set_frequencies('0', 0, frequencies0)

    frequencies1 = list(np.arange(100).reshape(1, -1) + 10)
    cm.set_frequencies('1', 10, frequencies1)

    positions = cm.get_position_by_time(0, [10])
    assert positions == [10]
    positions = cm.get_position_by_time(0, [10.5])
    assert positions == [10.5]
    positions = cm.get_position_by_time(5, [10])
    assert positions == [5]


def test_RayleighFitSet_average_fits():
    cm = CalibrationModel()

    cm.add_rayleigh_fit('0', 0, 0,
                        10, 1, 300, 100)
    cm.add_rayleigh_fit('0', 0, 1,
                        12, 1, 300, 100)

    w0_avg = cm.rayleigh_fits.average_fits('0', 0)

    assert w0_avg == 11.0


def test_BrillouinFitSet_average_fits():
    cm = CalibrationModel()

    cm.add_brillouin_fit('0', 0, 0,
                         10, 1, 300, 100)
    cm.add_brillouin_fit('0', 0, 1,
                         12, 1, 300, 100)

    w0_avg = cm.brillouin_fits.average_fits('0', 0)

    assert w0_avg == 11.0


def test_fitset():
    fit_set = FitSet()

    assert fit_set.make_key('1', 0, 0) == '1::0::0'
    assert fit_set.make_key('10', 0, 0) == '10::0::0'
    assert fit_set.split_key('1::0::0') == ['1', 0, 0]
    assert fit_set.split_key('10::0::0') == ['10', 0, 0]

    fit0 = BrillouinFit('1', 1, 4, 11., 12., 13., 14.)
    fit_set.add_fit(fit0)
    fit1 = BrillouinFit('10', 3, 4, 11., 12., 13., 14.)
    fit_set.add_fit(fit1)

    assert fit_set.get_fit('1', 1, 4) == fit0
    assert fit_set.get_fit('1', 2, 4) is None
    assert fit_set.get_fit('10', 3, 4) == fit1
    assert fit_set.get_fit('2', 3, 4) is None

    fit_set.clear('1')

    assert fit_set.get_fit('1', 1, 4) is None
    assert fit_set.get_fit('10', 3, 4) == fit1

    fit_set.clear('10')

    assert fit_set.get_fit('10', 3, 4) is None



import logging

import numpy as np
from scipy import interpolate

from bmlab.models.regions import regions_merge_add_region
from bmlab.serializer import Serializer

logger = logging.getLogger(__name__)


class CalibrationModel(Serializer):

    def __init__(self):
        self.calib_times = {}
        self.spectra = {}
        self.brillouin_regions = {}
        self.rayleigh_regions = {}
        self.brillouin_fits = BrillouinFitSet()
        self.rayleigh_fits = RayleighFitSet()

        self.vipa_params = {}
        self.frequencies = {}

        self.frequency_by_calib_key_interpolators = {}
        self.frequencies_by_time_interpolator = None
        self.frequency_by_time_interpolator = None

    def post_deserialize(self):
        self.refresh_frequency_interpolators()

    def add_brillouin_region(self, calib_key, region):
        if calib_key not in self.brillouin_regions:
            self.brillouin_regions[calib_key] = []

        region = tuple(round(x) for x in region)

        regions_merge_add_region(
            self.brillouin_regions[calib_key], region)

    def set_brillouin_region(self, calib_key, index, region):
        if calib_key not in self.brillouin_regions:
            self.brillouin_regions[calib_key] = []

        region = tuple(round(x) for x in region)

        if index < len(self.brillouin_regions[calib_key]):
            self.brillouin_regions[calib_key][index] = region
        else:
            self.brillouin_regions[calib_key].append(region)

    def get_brillouin_regions(self, calib_key):
        regions = self.brillouin_regions.get(calib_key)
        if regions is None:
            return []
        return regions

    def clear_brillouin_regions(self, calib_key):
        self.brillouin_regions[calib_key] = []

    def add_brillouin_fit(self, calib_key, region, frame_num,
                          w0s, fwhms, intensities, offset):
        fit = BrillouinFit(calib_key, region, frame_num,
                           w0s, fwhms, intensities, offset)
        self.brillouin_fits.add_fit(fit)

    def get_brillouin_fit(self, calib_key, region, frame_num):
        return self.brillouin_fits.get_fit(calib_key, region, frame_num)

    def clear_brillouin_fits(self, calib_key):
        self.brillouin_fits.clear(calib_key)

    def add_rayleigh_region(self, calib_key, region):
        if calib_key not in self.rayleigh_regions:
            self.rayleigh_regions[calib_key] = []

        region = tuple(round(x) for x in region)

        regions_merge_add_region(
            self.rayleigh_regions[calib_key], region)

    def set_rayleigh_region(self, calib_key, index, region):
        if calib_key not in self.rayleigh_regions:
            self.rayleigh_regions[calib_key] = []

        region = tuple(round(x) for x in region)

        if index < len(self.rayleigh_regions[calib_key]):
            self.rayleigh_regions[calib_key][index] = region
        else:
            self.rayleigh_regions[calib_key].append(region)

    def get_rayleigh_regions(self, calib_key):
        regions = self.rayleigh_regions.get(calib_key)
        if regions is None:
            return []
        return regions

    def clear_rayleigh_regions(self, calib_key):
        self.rayleigh_regions[calib_key] = []

    def add_rayleigh_fit(self, calib_key, region, frame_num,
                         w0, fwhm, intensity, offset):
        fit = RayleighFit(calib_key, region, frame_num,
                          w0, fwhm, intensity, offset)
        self.rayleigh_fits.add_fit(fit)

    def get_rayleigh_fit(self, calib_key, region, frame_num):
        return self.rayleigh_fits.get_fit(calib_key, region, frame_num)

    def clear_rayleigh_fits(self, calib_key):
        self.rayleigh_fits.clear(calib_key)

    def set_spectra(self, calib_key, spectra):
        self.spectra[calib_key] = spectra

    def get_spectra(self, calib_key):
        spectra = self.spectra.get(calib_key)
        if spectra:
            return spectra
        return None

    def get_sorted_peaks(self, calib_key, frame_num):
        """
        Returns the sorted centers of all fitted peaks
        of a given calibration and frame.

        Parameters
        ----------
        calib_key: The calibration key
        frame_num: The frame number

        Returns
        -------
        sorted np.array of all peaks
        """
        peaks = []
        # Search all fits for given calib_key and frame_num
        for key, fit in self.rayleigh_fits.fits.items():
            if (fit.calib_key == calib_key) and (fit.frame_num == frame_num):
                peaks.append(fit.w0)

        for key, fit in self.brillouin_fits.fits.items():
            if (fit.calib_key == calib_key) and (fit.frame_num == frame_num):
                for w0 in fit.w0s:
                    peaks.append(w0)

        return np.sort(np.array(peaks))

    def set_vipa_params(self, calib_key, vipa_params):
        self.vipa_params[calib_key] = vipa_params

    def clear_vipa_params(self, calib_key):
        if calib_key in self.vipa_params:
            del self.vipa_params[calib_key]

    def set_frequencies(self, calib_key, time, frequencies):
        self.frequencies[calib_key] = frequencies
        self.calib_times[calib_key] = time
        self.refresh_frequency_interpolators()

    def clear_frequencies(self, calib_key):
        if calib_key in self.frequencies:
            del self.frequencies[calib_key]
        if calib_key in self.calib_times:
            del self.calib_times[calib_key]
        self.refresh_frequency_interpolators()

    def refresh_frequency_interpolators(self):
        """
        This function creates the interpolator to get
        the frequencies of points in a spectrum

        Returns
        -------
        """
        # Reset all interpolators
        self.frequency_by_calib_key_interpolators = {}
        self.frequencies_by_time_interpolator = None
        self.frequency_by_time_interpolator = None

        sorted_keys = sorted(self.calib_times,
                             key=self.calib_times.get)
        # Don't do anything if there are not calibrations
        if len(sorted_keys) < 1:
            return

        """
        Create the interpolator for getting a frequency in a
        calibration spectrum
        """
        for calib_key in self.frequencies:
            frequencies = self.get_frequencies_by_calib_key(calib_key)
            if frequencies is None or not frequencies:
                return
            frequency = np.mean(np.array(frequencies), axis=0)

            xdata = np.arange(len(frequency))
            self.frequency_by_calib_key_interpolators[calib_key] =\
                interpolate.interp1d(xdata, frequency)

        """
        Create the interpolator for getting frequencies by time
        """
        if len(sorted_keys) < 2:
            self.frequencies_by_time_interpolator = \
                lambda time: np.nanmean(self.frequencies[sorted_keys[0]], 0)
        else:
            calib_times_array = []
            frequencies = []
            for key in sorted_keys:
                calib_times_array.append(self.calib_times[key])
                frequencies.append(np.nanmean(self.frequencies[key], 0))

            calib_times_array = np.array(calib_times_array)
            frequencies = np.squeeze(frequencies)

            self.frequencies_by_time_interpolator =\
                interpolate.interp1d(calib_times_array, frequencies, axis=0)

        """
        Create the interpolator for getting a frequency by position and time
        """
        calib_times_array = []
        frequencies = []
        for key in sorted_keys:
            calib_times_array.append(self.calib_times[key])
            frequencies.append(np.nanmean(self.frequencies[key], 0))

        calib_times_array = np.array(calib_times_array)
        frequencies = np.array(frequencies)

        indices = np.arange(frequencies.shape[1])

        # If we only have one time point, we
        # interpolate by peak position only
        if len(sorted_keys) < 2:
            frequencies = np.squeeze(frequencies)
            f = interpolate.interp1d(
                indices,
                frequencies,
                bounds_error=False
            )
            self.frequency_by_time_interpolator =\
                lambda time, position: f(position)
        # Otherwise we can interpolate by time as well
        else:
            f = interpolate.RegularGridInterpolator(
                (calib_times_array, indices),
                frequencies,
                method='linear',
                bounds_error=False
            )
            self.frequency_by_time_interpolator = \
                lambda time, position: f((time, position))

    def get_frequencies_by_calib_key(self, calib_key):
        """
        Returns the complete frequency axis for a given

        Parameters
        ----------
        calib_key: The key of the calibration

        Returns
        -------
        The frequency axis in Hz
        """
        if calib_key in self.frequencies:
            return self.frequencies[calib_key]

    def get_frequency_by_calib_key(self, position, calib_key):
        """
        Returns the frequency of a peak position on the
        spectrum for a given calibration

        Parameters
        ----------
        position: The position(s) of the peak(s)
        calib_key: The key of the calibration

        Returns
        -------
        The corresponding frequency in Hz
        """
        if calib_key in self.frequency_by_calib_key_interpolators:
            return self.frequency_by_calib_key_interpolators[
                calib_key](position)

    def get_frequencies_by_time(self, time):
        """
        Returns the complete frequency axis for a given

        Parameters
        ----------
        time: The time

        Returns
        -------
        The frequency axis in Hz
        """
        if self.frequencies_by_time_interpolator is not None:
            return self.frequencies_by_time_interpolator(time)

    def get_frequency_by_time(self, time, position):
        """
        Returns the frequency of a peak position on the
        spectrum for a given time

        Parameters
        ----------
        time: The time
        position: The position(s) of the peak(s)
        on the spectrum

        Returns
        -------
        The corresponding frequency in Hz
        """
        # Cannot execute if ndims are unequal
        if np.ndim(time) is not np.ndim(position):
            return None

        shape_time = np.shape(time)
        shape_position = np.shape(position)

        # In case the arrays don't have the same shape,
        # we repeat the time array
        if shape_time is not shape_position:
            time = np.tile(
                time,
                (np.array(shape_position) - np.array(shape_time)) + 1
            )

        if self.frequency_by_time_interpolator is not None:
            return self.frequency_by_time_interpolator(time, position)

    def get_position_by_time(self, time, frequencies):
        """
        Returns the position on the spectrum in pix for a given frequency
        at a certain time

        Parameters
        ----------
        time: The time
        frequencies: [GHz] The frequencies we want to convert to pixel

        Returns
        -------

        """
        spectrum = self.get_frequencies_by_time(time)
        f = interpolate.interp1d(spectrum, range(len(spectrum)))
        return f(frequencies)


class FitSet(Serializer):

    def __init__(self):
        self.fits = {}

    def make_key(self, calib_key, region_key, frame_num):
        return calib_key + '::' + str(region_key) + '::' + str(frame_num)

    def split_key(self, key):
        items = key.split('::')
        items[1] = int(items[1])
        items[2] = int(items[2])
        return items

    def add_fit(self, fit):
        key = self.make_key(fit.calib_key, fit.region_key, fit.frame_num)
        self.fits[key] = fit

    def get_fit(self, calib_key, region_key, frame_num=None):
        key = self.make_key(calib_key, region_key, frame_num)
        return self.fits.get(key)

    def clear(self, calib_key):
        keys = []
        for key, value in self.fits.items():
            if calib_key == self.split_key(key)[0]:
                keys.append(key)
        for key in keys:
            del self.fits[key]


class RayleighFitSet(FitSet, Serializer):

    def average_fits(self, calib_key, region_key):
        w0s = []
        for key, fit in self.fits.items():
            calib_key_, region_key_, _ = self.split_key(key)
            if calib_key == calib_key_ and region_key == region_key_:
                w0s.append(fit.w0)
        logger.debug('w0s = ', w0s)
        if w0s:
            return np.mean(w0s)
        return None


class BrillouinFitSet(FitSet, Serializer):

    def average_fits(self, calib_key, region_key):
        w0s = []
        for key, fit in self.fits.items():
            calib_key_, region_key_, _ = self.split_key(key)
            if calib_key == calib_key_ and region_key == region_key_:
                w0s.append(fit.w0s)
        logger.debug('w0s = ', w0s)
        if w0s:
            w0s = np.array(w0s)
            return np.mean(w0s, axis=0)
        return None


class RayleighFit(Serializer):

    def __init__(self, calib_key, region_key, frame_num,
                 w0, fwhm, intensity, offset):
        self.calib_key = calib_key
        self.region_key = region_key
        self.frame_num = frame_num
        self.w0 = w0
        self.fwhm = fwhm
        self.intensity = intensity
        self.offset = offset


class BrillouinFit(Serializer):

    def __init__(self, calib_key, region_key, frame_num,
                 w0s, fwhms, intensities, offset):
        self.calib_key = calib_key
        self.region_key = region_key
        self.frame_num = frame_num
        self.w0s = w0s
        self.fwhms = fwhms
        self.intensities = intensities
        self.offset = offset



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import exists, dirname, realpath
from setuptools import setup, find_packages
import sys

author = u"bmlab developers"
# authors in alphabetical order
authors = [
    "Matthias Bär",
    "Paul Müller",
    "Raimund Schlüßler",
    "Timon Beck",
]
description = 'Library for Brillouin microscopy data analysis'
name = 'bmlab'
year = "2022"


sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
try:
    from _version import version
except BaseException:
    version = "unknown"


setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/BrillouinMicroscopy/bmlab',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="GPL v3",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["h5py>=2.10.0",
                      "matplotlib",
                      "numpy>=1.17.0",
                      "packaging>=20.8",
                      "pillow",
                      "pytest",
                      "pytest_mock",
                      "scikit-image>=0.19.0",
                      "scipy<1.9.0",
                      "shapely>=1.8.2"
                      ],
    # not to be confused with definitions in pyproject.toml [build-system]
    python_requires=">=3.7",
    keywords=["Brillouin microscopy"],
    classifiers=['Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 'Topic :: Scientific/Engineering :: Visualization',
                 'Intended Audience :: Science/Research',
                 ],
    platforms=['ALL'],
)



import pathlib

import numpy as np

from bmlab.fits import lorentz, fit_lorentz, fit_circle,\
    fit_double_lorentz, calculate_exact_circle, fit_vipa, VIPA,\
    are_points_on_line

from bmlab.models.setup import AVAILABLE_SETUPS

import pytest


def test_fit_lorentz():
    # Arrange
    x = np.linspace(0, 30, 100)
    w0 = 15.
    fwhm = 4
    offset = 10.
    intensity = 10.
    y_data = lorentz(x, w0, fwhm, intensity) + offset

    actual_w0, actual_fwhm, actual_intensity, actual_offset =\
        fit_lorentz(x, y_data)
    np.testing.assert_almost_equal(actual_w0, w0, decimal=3)
    np.testing.assert_almost_equal(actual_fwhm, fwhm, decimal=3)
    np.testing.assert_almost_equal(actual_intensity, intensity, decimal=3)
    np.testing.assert_almost_equal(actual_offset, offset, decimal=3)


def test_fit_lorentz_real_image_data():
    """ The data for this test case has been extracted manually
        from the running BMicro application.
    """

    data_dir = pathlib.Path(__file__).parent / 'data'

    region = np.load(data_dir / 'rayleigh_reg0_region.npy')
    xdata = np.load(data_dir / 'rayleigh_reg0_xdata.npy')
    ydata = np.load(data_dir / 'rayleigh_reg0_ydata.npy')

    w0, fwhm, intensity, offset = fit_lorentz(xdata[range(*region)],
                                              ydata[range(*region)])

    assert w0 == pytest.approx(115, 0.2)
    assert fwhm == pytest.approx(6, 0.5)
    assert intensity == pytest.approx(1186, 1)
    assert offset == pytest.approx(47, 1)


def test_fit_double_lorentz():
    # Arrange
    x = np.linspace(0, 60, 200)
    w0_left = 10.
    intensity_left = 8.
    fwhm_left = 2.
    w0_right = 20.
    intensity_right = 12.
    fwhm_right = 4.
    offset = 10.
    y_data = lorentz(x, w0_left, fwhm_left, intensity_left)
    y_data += lorentz(x, w0_right, fwhm_right, intensity_right) + offset

    w0s, fwhms, intens, actual_offset\
        = fit_double_lorentz(x, y_data)

    np.testing.assert_almost_equal(w0s[0], w0_left, decimal=3)
    np.testing.assert_almost_equal(fwhms[0], fwhm_left, decimal=3)
    np.testing.assert_almost_equal(intens[0], intensity_left, decimal=3)

    np.testing.assert_almost_equal(w0s[1], w0_right, decimal=3)
    np.testing.assert_almost_equal(fwhms[1], fwhm_right, decimal=3)
    np.testing.assert_almost_equal(intens[1], intensity_right, decimal=3)

    np.testing.assert_almost_equal(actual_offset, offset, decimal=3)


def test_fit_double_lorentz_with_bounds():
    # Arrange
    x = np.linspace(0, 60, 500)
    w0_left = 20.
    intensity_left = 8.
    fwhm_left = 4.0
    w0_right = 40.
    intensity_right = 12.
    fwhm_right = 5.0
    offset = 10.
    y_data = lorentz(x, w0_left, fwhm_left, intensity_left)
    y_data += lorentz(x, w0_right, fwhm_right, intensity_right) + offset

    bounds = ((19, 19.9), (-np.Inf, np.Inf))

    w0s, fwhms, intens, actual_offset \
        = fit_double_lorentz(x, y_data, bounds)

    np.testing.assert_almost_equal(w0s[0], 19.9, decimal=2)
    np.testing.assert_almost_equal(fwhms[0], fwhm_left, decimal=1)
    np.testing.assert_almost_equal(intens[0], intensity_left, decimal=1)

    np.testing.assert_almost_equal(w0s[1], w0_right, decimal=1)
    np.testing.assert_almost_equal(fwhms[1], fwhm_right, decimal=1)
    np.testing.assert_almost_equal(intens[1], intensity_right, decimal=1)

    np.testing.assert_almost_equal(actual_offset, offset, decimal=1)

    bounds = ((-np.Inf, np.Inf), (20.1, 21))

    w0s, fwhms, intens, actual_offset \
        = fit_double_lorentz(x, y_data, bounds)

    np.testing.assert_almost_equal(w0s[0], w0_right, decimal=1)
    np.testing.assert_almost_equal(fwhms[0], fwhm_right, decimal=1)
    np.testing.assert_almost_equal(intens[0], intensity_right, decimal=1)

    np.testing.assert_almost_equal(w0s[1], 20.1, decimal=1)
    np.testing.assert_almost_equal(fwhms[1], fwhm_left, decimal=1)
    np.testing.assert_almost_equal(intens[1], intensity_left, decimal=1)

    np.testing.assert_almost_equal(actual_offset, offset, decimal=1)


def test_circle_fit():
    expect_r = 550
    expect_c = (-200, -220)
    n_test_data_points = 6
    noise_strength = 15
    np.random.seed(1)
    x_noise = np.random.random(n_test_data_points) * noise_strength
    y_noise = np.random.random(n_test_data_points) * noise_strength

    test_points_noise = [(expect_r * np.cos(phi) + x_noise[i] + expect_c[0],
                          expect_r * np.sin(phi) + y_noise[i] + expect_c[1])
                         for i, phi in enumerate(
        np.linspace(0.1, np.pi / 2, n_test_data_points))
    ]

    actual_c_noise, actual_r_noise = fit_circle(test_points_noise)

    np.testing.assert_allclose(actual_c_noise, expect_c, rtol=0.05)
    np.testing.assert_allclose(actual_r_noise, expect_r, rtol=0.05)

    test_points = [(expect_r * np.cos(phi) + 0 * x_noise[i] + expect_c[0],
                    expect_r * np.sin(phi) + 0 * y_noise[i] + expect_c[1])
                   for i, phi in enumerate(
        np.linspace(0.1, np.pi / 2, int(n_test_data_points / 2) + 1))
    ]

    actual_c, actual_r = fit_circle(test_points)
    np.testing.assert_allclose(actual_c, expect_c, rtol=1e-2)
    np.testing.assert_allclose(actual_r, expect_r, rtol=1e-2)

    # Test that radius is always positive
    test_points_2 = [(np.sqrt(2), 0), (1, 1), (0, np.sqrt(2))]
    actual_c_2, actual_r_2 = fit_circle(test_points_2)

    np.testing.assert_allclose(actual_c_2, (0, 0), rtol=0.001, atol=0.0001)
    assert (actual_r_2 > 0)
    np.testing.assert_allclose(actual_r_2, np.sqrt(2),
                               rtol=0.0001, atol=0.000001)

    test_points_3 = [
        (10.0, 0.0),
        (7.0710678118654755, 7.0710678118654755),
        (6.123233995736766e-16, 10.0)
    ]
    actual_c_3, actual_r_3 = fit_circle(test_points_3)

    np.testing.assert_allclose(actual_c_3, (0, 0), rtol=0.001, atol=0.0001)
    assert (actual_r_3 > 0)
    np.testing.assert_allclose(actual_r_3, 10, rtol=0.0001, atol=0.000001)


def test_calculate_exact_circle():
    points = [[1, 0], [0, 1], [-1, 0]]
    center, radius = calculate_exact_circle(points)

    np.testing.assert_allclose(center, [0, 0], rtol=1e-2)
    np.testing.assert_allclose(radius, 1, rtol=1e-2)

    points = [[-1, -2], [-2, -1], [-3, -2]]
    center, radius = calculate_exact_circle(points)

    np.testing.assert_allclose(center, [-2, -2], rtol=1e-2)
    np.testing.assert_allclose(radius, 1, rtol=1e-2)

    points = [[8, -2], [-2, 8], [-12, -2]]
    center, radius = calculate_exact_circle(points)

    np.testing.assert_allclose(center, [-2, -2], rtol=1e-2)
    np.testing.assert_allclose(radius, 10, rtol=1e-2)


def test_are_points_on_line():
    assert are_points_on_line([(0, 0), (1, 1), (2, 2)])
    assert are_points_on_line([(0, 0), (1, 1)])
    assert not are_points_on_line([(0, 0), (1, 1), (3, 2)])
    assert not are_points_on_line([(0, 0), (1, 1), (3, 2), (4, 4)])
    assert are_points_on_line([(0, 0), (1, 1), (2, 2), (3, 4)])


def test_fit_vipa():
    setup = AVAILABLE_SETUPS[0]
    peaks = np.array([
        84.2957567375179,
        147.651886970066,
        166.035559534916,
        229.678232333124,
        244.118149639528,
        287.316083765756
    ])

    vipa_params = fit_vipa(peaks, setup)

    # Values extracted from the previous Matlab version
    vipa_params_expected =\
        np.array([
            2.602626743299098e-15, -2.565391389072813e-22,
            -6.473779072623057e-25, 14.89190812511701e+9
        ])

    actual = VIPA(peaks, vipa_params) - setup.f0
    expected = VIPA(peaks, vipa_params_expected) - setup.f0

    # Values are in GHz, differences below 5 MHz should be good
    np.testing.assert_allclose(actual, expected, atol=5e3)



from bmlab.models.evaluation_model import EvaluationModel


def test_initialize_results_arrays():
    evm = EvaluationModel()
    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 1,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 1,
    })

    assert evm.results['intensity'].shape == (5, 5, 5, 2, 1, 1)
    assert evm.results['brillouin_peak_position'].shape == (5, 5, 5, 2, 1, 1)
    assert evm.results['rayleigh_peak_position'].shape == (5, 5, 5, 2, 1, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })

    assert evm.results['intensity'].shape == (5, 5, 5, 2, 1, 1)
    assert evm.results['brillouin_peak_position'].shape == (5, 5, 5, 2, 2, 1)
    assert evm.results['rayleigh_peak_position'].shape == (5, 5, 5, 2, 2, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 2,
        'nr_rayleigh_regions': 2,
    })

    assert evm.results['intensity'].shape == (5, 5, 5, 2, 1, 1)
    # If we have more than one Brillouin peak,
    # we store the result of a single peak fit
    # and the result of a multi peak fit, adding up to nr_brillouin_peaks + 1
    assert evm.results['brillouin_peak_position'].shape == (5, 5, 5, 2, 2, 3)
    assert evm.results['rayleigh_peak_position'].shape == (5, 5, 5, 2, 2, 1)



import os
import errno

from pathlib import Path

import h5py
from numpy import transpose
import math

from bmlab import __version__ as version
from bmlab.file import BrillouinFile, is_source_file, is_session_file
from bmlab.models.extraction_model import ExtractionModel
from bmlab.models.orientation import Orientation
from bmlab.models.calibration_model import CalibrationModel
from bmlab.models.peak_selection_model import PeakSelectionModel
from bmlab.models.evaluation_model import EvaluationModel
from bmlab.serializer import Serializer


def get_session_file_path(source_file, create_folder=False):
    # If the raw data file is located in a 'RawData' folder,
    # we put the eval data file in an 'EvalData' folder and
    # don't append the 'session' string.
    file = Path(source_file).resolve()
    if file.parent.name == 'RawData':
        eval_folder = file.parents[1] / 'EvalData'
        # Create the evaluation folder if necessary
        if create_folder and not os.path.exists(eval_folder):
            os.mkdir(eval_folder)
        return Path(str(eval_folder / (str(file.name)[:-3] + '.h5')))
    else:
        return Path(str(source_file)[:-3] + '.session.h5')


def get_source_file_path(session_file):
    # If the session file is located in a 'EvalData' folder,
    # we find the source file in an 'RawData' folder
    file = Path(session_file).resolve()
    if file.parent.name == 'EvalData':
        raw_folder = file.parents[1] / 'RawData'
        return Path(str(raw_folder / (str(file.name)[:-3] + '.h5')))
    else:
        return Path(str(session_file)[:-11] + '.h5')


def get_valid_source(path):
    path = Path(path)
    # Check whether this file exists at all
    if not os.path.exists(path):
        raise FileNotFoundError(
            errno.ENOENT,
            "The file '{}' does not exist."
            .format(path),
            path
        )

    # If this is a session file, we need to check whether
    # the source file exists and is valid.
    if is_session_file(path):
        source_file_path = get_source_file_path(path)
        # If the session file is valid, but the source file
        # does not exist, we raise a file not found error.
        if not os.path.exists(source_file_path):
            raise FileNotFoundError(
                errno.ENOENT,
                "Could not find the corresponding source"
                " data file '{}' for session file '{}'."
                .format(source_file_path, path)
                + " Please ensure the source data file exists.",
                source_file_path
            )
        if is_source_file(source_file_path):
            return source_file_path
        # If the session file is valid, but the source file
        # is not, we raise a BmlabInvalidFileError
        else:
            raise BmlabInvalidFileError(
                errno.ENOENT,
                "Could not open file '{}.".format(source_file_path)
                + " The file is not a valid BrillouinAcquisition file.",
                source_file_path
            )

    # If this is a source file, just return its path
    if is_source_file(path):
        return path

    # If we end up here,
    # the file provided was neither a bmlab session file
    # nor a raw data file from BrillouinAcquisition.
    raise BmlabInvalidFileError(
        errno.ENOENT,
        "Could not open file '{}'.".format(path)
        + " The provided file is neither a valid"
          " BrillouinAcquisition nor bmlab file.",
        path
    )


class BmlabInvalidFileError(FileNotFoundError):
    def __init__(self, *args, **kwargs):
        super(BmlabInvalidFileError, self).__init__(*args, **kwargs)


class Session(Serializer):
    """
    Session stores information about the current file
    to be processed.

    Session is a singleton. It can be accessed by
    calling the get_instance() method.
    """

    __instance = None

    def __init__(self):
        """
        Constructor of Session class. Since Session is a singleton,
        users should not call it but instead use the get_instance
        method.
        """
        if Session.__instance is not None:
            raise Exception('Session is a singleton!')
        else:
            Session.__instance = self
            self.clear()

    def current_repetition(self):
        """
        Returns the repetition currently selected in data tab
        """
        if self.file and self._current_repetition_key:
            return self.file.get_repetition(self._current_repetition_key)
        return None

    def set_current_repetition(self, rep_key):
        self._current_repetition_key = rep_key

    def extraction_model(self):
        """
        Returns ExtractionModel instance for currently selected repetition
        """
        return self.extraction_models.get(self._current_repetition_key)

    def calibration_model(self):
        return self.calibration_models.get(self._current_repetition_key)

    def peak_selection_model(self):
        return self.peak_selection_models.get(self._current_repetition_key)

    def evaluation_model(self):
        return self.evaluation_models.get(self._current_repetition_key)

    def set_image_shape(self):
        """
        We set the extraction model image shape for every repetition when
        - a file gets loaded
        - the image orientation changes
        """
        if self.file is None:
            return

        repetitions = self.file.repetition_keys()
        for repetition in repetitions:
            imgs = self.file.get_repetition(repetition).payload.get_image('0')
            # If no images are available, skip this repetition
            if imgs is None:
                continue
            img = self.orientation.apply(imgs[0, ...])
            self.extraction_models.get(repetition).set_image_shape(img.shape)

    def set_arc_width(self):
        """
        We set the arc_width for arc extraction on file load.
        """
        if self.file is None:
            return

        repetitions = self.file.repetition_keys()
        for repetition in repetitions:
            binning_factor = self.file.get_repetition(repetition)\
                .payload.get_binning_factor('0')

            em = self.extraction_models.get(repetition)
            arc_width = math.ceil(em.arc_width / binning_factor)
            em.set_arc_width(arc_width)

    @staticmethod
    def get_instance():
        """
        Returns the singleton instance of Session

        Returns
        -------
        out: Session
        """
        if Session.__instance is None:
            Session()
        return Session.__instance

    def set_file(self, file_name):
        """
        Set the file to be processed.

        Loads the corresponding data from HDF file.

        Parameters
        ----------
        file_name : Path
            The file name.

        """
        # Get the source file in case it's a session file
        file_name = get_valid_source(file_name)
        # There is no valid source file
        if file_name is None:
            raise Exception('No source data file found')

        try:
            file = BrillouinFile(file_name)
        except Exception as e:
            raise e
        else:
            """ Only load data if the file could be opened """
            session = Session.get_instance()
            session.file = file
            session.extraction_models = {
                key: ExtractionModel()
                for key in self.file.repetition_keys()
            }
            session.calibration_models = {
                key: CalibrationModel()
                for key in self.file.repetition_keys()
            }
            session.peak_selection_models = {
                key: PeakSelectionModel()
                for key in self.file.repetition_keys()
            }
            session.evaluation_models = {
                key: EvaluationModel()
                for key in self.file.repetition_keys()
            }
            self.set_image_shape()
            self.set_arc_width()

        try:
            self.load(file_name)
            self.set_arc_width()
        except Exception as e:
            raise e
        else:
            Session.get_instance().file = file

    def get_calib_keys(self, sort_by_time=False):
        if self.current_repetition() is None:
            return None
        return self.current_repetition()\
            .calibration.image_keys(sort_by_time=sort_by_time)

    def get_calibration_image(self, calib_key, frame_num=None):
        if self.current_repetition() is None:
            return None
        imgs = self.current_repetition().calibration.get_image(calib_key)
        if frame_num is not None:
            imgs = imgs[frame_num, ...]
        return self.orientation.apply(imgs)

    def get_calibration_image_count(self, calib_key):
        return self.current_repetition()\
            .calibration.get_image_count(calib_key)

    def get_calibration_time(self, calib_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().calibration.get_time(calib_key)

    def get_calibration_exposure(self, calib_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().calibration.get_exposure(calib_key)

    def get_calibration_binning(self, calib_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().calibration.get_binning(calib_key)

    def get_calibration_binning_factor(self, calib_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition()\
            .calibration.get_binning_factor(calib_key)

    def get_image_keys(self, sort_by_time=False):
        if self.current_repetition() is None:
            return None
        return self.current_repetition()\
            .payload.image_keys(sort_by_time=sort_by_time)

    def get_payload_image(self, image_key, frame_num=None):
        if self.current_repetition() is None:
            return None
        imgs = self.current_repetition().payload.get_image(image_key)
        if frame_num is not None:
            imgs = imgs[frame_num, ...]
        return self.orientation.apply(imgs)

    def get_payload_image_count(self, calib_key):
        return self.current_repetition().payload.get_image_count(calib_key)

    def get_payload_time(self, image_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().payload.get_time(image_key)

    def get_payload_exposure(self, image_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().payload.get_exposure(image_key)

    def get_payload_binning(self, image_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().payload.get_binning(image_key)

    def get_payload_binning_factor(self, image_key):
        if self.current_repetition() is None:
            return None
        return self.current_repetition()\
            .payload.get_binning_factor(image_key)

    def get_payload_resolution(self):
        if self.current_repetition() is None:
            return None
        return self.current_repetition().payload.resolution

    def get_payload_positions(self):
        if self.current_repetition() is None:
            return None
        positions = self.current_repetition().payload.positions
        # We need to correctly transpose the array to have the
        # axes in order x-y-z
        for axis in positions:
            positions[axis] = transpose(positions[axis], axes=(1, 2, 0))
        return positions

    def clear(self):
        """
        Close connection to loaded file.
        """

        # Global session data:
        if hasattr(self, 'file') and self.file is not None:
            self.file.close()
            self.file = None
        else:
            self.file = None

        self.orientation = Orientation()
        self.setup = None

        # Session data by repetition:
        self.extraction_models = {}
        self.calibration_models = {}
        self.evaluation_models = {}
        self.peak_selection_models = {}

        self._current_repetition_key = None

    def set_setup(self, setup):
        self.setup = setup

    def set_rotation(self, num_rots):
        self.orientation.set_rotation(num_rots)
        self.set_image_shape()

    def set_reflection(self, **kwargs):
        self.orientation.set_reflection(**kwargs)
        self.set_image_shape()

    def save(self):
        if self.file is None:
            return

        session_file_name = get_session_file_path(self.file.path,
                                                  create_folder=True)

        with h5py.File(session_file_name, 'w') as f:
            self.serialize(f, 'session', skip=['file'])
            # Store the current bmlab version
            f.attrs['version'] = 'bmlab_' + version

    def load(self, h5_file_name):

        session_file_name = get_session_file_path(h5_file_name)

        if not os.path.exists(session_file_name):
            return

        if not is_session_file(session_file_name):
            raise BmlabInvalidFileError(
                errno.ENOENT,
                "Could not load the session file '{}'."
                .format(session_file_name)
                + " Please ensure the session file is valid.",
                session_file_name
            )

        with h5py.File(session_file_name, 'r') as f:
            new_session = Serializer.deserialize(f['session'])
            session = Session.get_instance()
            for var_name, var_value in new_session.__dict__.items():
                session.__dict__[var_name] = var_value



import numpy as np

from bmlab.models.extraction_model import ExtractionModel
from bmlab.geometry import Circle


def test_add_and_set_points():
    em = ExtractionModel()
    em.add_point('0', 0, 0, 1)

    assert (0, 1) in em.get_points('0')

    em.set_point('0', 0, 0, 0, 2)
    em.set_point('0', 2, 0, 1, 2)

    assert em.get_points('0')[0] == (0, 2)
    assert em.get_points('0')[-1] == (1, 2)


def test_extraction_model_triggers_position_interpolation():

    em = ExtractionModel()

    em.set_image_shape((500, 500))
    em.add_point('0', 0, 0, 1)
    em.add_point('0', 0, 1, 0)

    assert (0, 1) in em.get_points('0')
    assert (1, 0) in em.get_points('0')
    np.testing.assert_equal(em.get_arc_by_time(0), np.empty(0))

    em.add_point('0', 0, 0, -1)

    assert (0, -1) in em.get_points('0')
    assert em.get_arc_by_time(0).shape == (500, 5, 2)


def test_clear_points_from_extraction_model():
    em = ExtractionModel()
    em.add_point('0', 0, 15, 15)
    em.add_point('0', 0, 75, 85)

    em.clear_points('0')
    assert em.get_points('0') == []
    assert em.get_time('0') == []


def test_set_arc_width():
    em = ExtractionModel()

    # Test default width
    assert em.arc_width == 2

    em.set_arc_width(3)

    assert em.arc_width == 3


def test_get_arc_by_calib_key():
    calib_keys = ['0', '1']
    times = [100, 500]
    circle_center = [(0, 0), (10, 0)]
    circle_radii = [9, 10]
    arc_start_end = [
        [[0.0, 9.0], [9.0, 0.0]],
        [[0.0, 0.0], [20.0, 0.0]]
    ]

    em = ExtractionModel()
    em.set_image_shape((100, 100))
    em.set_arc_width(2)
    # Set the circle points and test getting arc by calib_key
    for i, calib_key in enumerate(calib_keys):
        radius = circle_radii[i]

        circle = Circle(circle_center[i], radius)

        phis = [0, np.pi/4, np.pi/2]
        for phi in phis:
            (xdata, ydata), _ = circle.point(phi)
            em.add_point(calib_key, times[i], xdata, ydata)

        # Get the arc
        arc = em.get_arc_by_calib_key(calib_key)

        assert len(arc) == 500

        np.testing.assert_allclose(arc[0, 2, :], arc_start_end[i][0],
                                   rtol=0.05, atol=0.000001)
        np.testing.assert_allclose(arc[-1, 2, :], arc_start_end[i][1],
                                   rtol=0.05, atol=0.000001)


def test_get_arc_by_time():
    """
    Test getting arc by time
    """
    calib_keys = ['0', '1']
    times = [100, 500]
    circle_center = [(0, 0), (0, 1)]
    circle_radii = [10, 10]

    em = ExtractionModel()
    em.set_image_shape((100, 100))
    em.set_arc_width(2)
    # Set the circle points and test getting arc by calib_key
    for i, calib_key in enumerate(calib_keys):
        radius = circle_radii[i]

        circle = Circle(circle_center[i], radius)

        phis = [0, np.pi/4, np.pi/2]
        for phi in phis:
            (xdata, ydata), _ = circle.point(phi)
            em.add_point(calib_key, times[i], xdata, ydata)

    # the exact time point of a calibration should yield
    # the same as the calibration key
    arc = em.get_arc_by_time(times[0])
    expected_arc = em.get_arc_by_calib_key(calib_keys[0])
    np.testing.assert_allclose(arc, expected_arc,
                               rtol=0.05, atol=0.000001)

    # a time point after the last calibration should give
    # the value of the last calibration
    arc = em.get_arc_by_time(-500)
    expected_arc = em.get_arc_by_calib_key(calib_keys[0])
    np.testing.assert_allclose(arc, expected_arc,
                               rtol=0.05, atol=0.000001)
    arc = em.get_arc_by_time(times[-1] + 500)
    expected_arc = em.get_arc_by_calib_key(calib_keys[-1])
    np.testing.assert_allclose(arc, expected_arc,
                               rtol=0.05, atol=0.000001)

    # a time point between two calibrations should give
    # an interpolated value
    arc = em.get_arc_by_time(np.mean(times[0:2]))
    expected_arc_start = [0.0, 10.5]
    np.testing.assert_allclose(arc[0, 2, :], expected_arc_start,
                               rtol=0.05, atol=0.000001)


def test_get_arc_from_circle_phis():
    circle = Circle((0, 0), 100)
    phis = [0, np.pi/2, np.pi]
    arc_width = 2

    arc_expected = np.array([
        [[98.0, 0.0], [99, 0.0], [100.0, 0.0], [101.0, 0.0], [102.0, 0.0]],
        [[0.0, 98.0], [0.0, 99.0], [0.0, 100.0], [0.0, 101.0], [0.0, 102.0]],
        [[-98.0, 0.0], [-99, 0.0], [-100.0, 0.0], [-101.0, 0.0], [-102.0, 0.0]]
    ])

    arc = ExtractionModel().get_arc_from_circle_phis(circle, phis, arc_width)
    np.testing.assert_allclose(arc, arc_expected, atol=1e-12)



import numpy as np
import math
from shapely.geometry import Polygon, Point
from bmlab.utils import debug_timer


class Circle(object):
    """
    Representation of a circle.

    Wraps shapely-API for reduced complexity.
    """

    def __init__(self, center, radius):
        """
        Creates a circle instance.

        Parameters
        ----------
        center: 2-tuple
            Center of the circle.

        radius: float
            Radius of the circle.

        """
        self.center = np.array(center)
        self.radius = radius

        # Check that we have a valid circle
        # - no property is Inf or NaN
        # - radius is not zero
        self.valid = True
        if radius == 0 or\
            not np.isfinite(center).all() or\
                not np.isfinite(radius):
            self.valid = False

    def is_valid(self):
        return self.valid

    def point(self, phi, integer=False):
        """
        Returns the xy-coordinates of a point with given polar angle phi.

        Parameters
        ----------
        phi: float
            The polar angle (counted from 0-axis to given point).

        integer: bool
            Flag indicating whether the result should be integer
            pixel value or floating point values.

        Returns
        -------
        point: 2-tuple
            The xy-coordinates of the point.

        """
        if not self.valid:
            return None
        e_r = self.e_r(phi)
        pt = self.center + self.radius * e_r
        if integer:
            pt[0] = round(pt[0])
            pt[1] = round(pt[1])
            return np.array(pt, dtype=np.int)
        return pt, e_r

    def intersection(self, rect):
        """
        Calculates the intersection of the circle with a Rectangle object.

        Parameters
        ----------
        rect: Rectangle
            the rectangle

        Returns
        -------
        intersection: list of 2-tuples
            Points of intersection.
        """
        if not self.valid:
            return None
        center = Point(self.center)
        inter = center.buffer(self.radius).boundary.intersection(
            rect.poly).boundary
        return [(p.x, p.y) for p in inter.geoms]

    def angle(self, point):
        """
        Returns the polar angle for a given point.

        Parameters
        ----------
        point: 2-tuple
            The point for which to calculate the polar angle.

        Returns
        -------
        angle: float
            The polar angle.
        """
        if not self.valid:
            return None
        delta = np.array(point) - self.center
        angle = math.atan2(delta[1], delta[0])
        if angle < 0:
            angle = angle + 2 * np.pi
        return angle

    @staticmethod
    def e_r(phi):
        return np.array([np.cos(phi), np.sin(phi)], dtype=float)


class Rectangle(object):
    def __init__(self, shape):
        self.poly = Polygon(
            [(0, 0), (0, shape[1]), (shape[0], shape[1]), (shape[0], 0)])


@debug_timer
def discretize_arc(circle, img_shape, num_points):
    """
    Returns a list of equidistant (in polar angle) points along a circle
    intersecting an image.

    Parameters
    ----------
    circle: Circle
        Circle that intersects the image.
    img_shape: 2-tuple
        Shape of the image.
    num_points: int
        How many points along the circle.

    Returns
    -------
    phis: numpy.ndarray
        polar angles of the discrete points on circular arc
    """
    if num_points is None:
        num_points = 2*(sum(np.square(img_shape)))**0.5
    rect = Rectangle(img_shape)
    cut_edges = circle.intersection(rect)
    cut_edges.sort(key=lambda edge: edge[0])
    if not cut_edges:
        return None
    phis_edges = [circle.angle(p) for p in cut_edges]
    if not phis_edges:
        return None
    return np.linspace(phis_edges[0], phis_edges[-1], num_points)
